/*
 * http://www.wayforward.net/avGPS/libav/av_comm.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 * ---
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#include <poll.h>

#include "av_comm.h"
#include "av_err.h"
#include "av_str.h"

static const tchar_t OPEN_FN[] = T("open");
static const tchar_t TCGETATTR_FN[] = T("tcgetattr");
static const tchar_t TCSETATTR_FN[] = T("tcsetattr");
static const tchar_t READ_FN[] = T("read");

static const tchar_t SOCKET_FN[] = T("socket");
static const tchar_t CONNECT_FN[] = T("connect");
static const tchar_t GETHOSTBYNAME_FN[] = T("gethostbyname");

static const tchar_t CFSETISPEED_FN[] = T("cfsetispeed");
static const tchar_t CFSETOSPEED_FN[] = T("cfsetospeed");
static const tchar_t MALLOC_FN[] = T("malloc");

/* on some systems, B4800 == 4800, but not on linux, sadly */
static const struct {
	unsigned baud;
	speed_t  constant;
} SPEEDS[] = {
	{ 4800, B4800 },
	{ 9600, B9600 },
	{ 19200, B19200 }
};

static int set_tty_fd(int fd, struct comm_device *pd, int baud,
                      const tchar_t *devicename);

/**
 * Open a comm device.  Sets baud rate, bits, parity, and stop bits.
 *
 * The _devicename_ can be a TCP/IP hostname prefixed by 'socket:'
 *
 * For example: 'socket:localhost:5000' to connect to TCP port 5000
 * on the local machine, or 'socket:gps.wayforward.net' to connect
 * to port 2947 (the default) on the 'gps.wayforward.net' machine.
 *
 * The _devicename_ can also be a serial device, like '/dev/gps' or
 * '/dev/tty0'.  If so, the _pc_ comm_settings structure is used to
 * set parity, bits, and baud rate.
 *
 * Otherwise, the _devicename_ is assumed to be a regular file.
 *
 * Returns 1 on success
 * Returns 0 on error, with error information pushed onto the "av_err.h"
 * error stack.
 */
int
comm_open(const char *devicename, const struct comm_settings *pc,
	  struct comm_device *pd) {
	int fd;

	if (memcmp(devicename, "socket:", 7) == 0) {
		struct sockaddr_in sin;
		tchar_t *hostname = tstr_save(devicename + 7), *portstr;

		if (hostname == NULL) {
			err_push_errno(MALLOC_FN, NULL);
			return 0;
		}

		portstr = tstrchr(hostname, ':');

		sin.sin_family = AF_INET;

		if (portstr != NULL) {
			*portstr++ = T('\0');
			sin.sin_port = htons(atoi(portstr));
		}
		else
			sin.sin_port = htons(2947);

		fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (fd < 0) {
			err_push_errno(SOCKET_FN, devicename + 7);
			free(hostname);
			return 0;
		}

		sin.sin_addr.s_addr = inet_addr(hostname);
		if (sin.sin_addr.s_addr == (in_addr_t) -1) {
			struct hostent *phe = gethostbyname(hostname);

			if (phe == NULL) {
				err_push_errno(GETHOSTBYNAME_FN, hostname);
				free(hostname);
				(void) close(fd);
				return 0;
			}

			sin.sin_addr = *(struct in_addr *) phe->h_addr;
		}

		free(hostname);

		if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
			err_push_errno(CONNECT_FN, devicename + 7);
			(void) close(fd);
			return 0;
		}

		pd->fd = fd;
		pd->type = comm_socket;
		pd->running = 1;
		return 1;
	}

	fd = open(devicename, O_RDONLY | O_NONBLOCK);

	if (fd < 0) {
		err_push_errno(OPEN_FN, devicename);
		return 0;
	}

	/*
	 * pulled from coreutils-5.97/src/stty.c
	 *
	 * We're going to compare the serial states to see
	 * if tcsetattr set all the important information.  To
	 * do that consistently, we must zero out any fields that
	 * tcgetattr may not set.
	 */
	memset(&pd->oldState, 0, sizeof(pd->oldState));

	if (!isatty(fd)) {
		pd->fd = fd;
		pd->running = 1;
		pd->type = comm_file;
		return 1;
	}
	else if (tcgetattr(fd, &pd->oldState) < 0) {
		err_push_errno(TCGETATTR_FN, devicename);
	}
	else if (set_tty_fd(fd, pd, pc->baud, devicename)) {
		return 1;
	}

	(void) close(fd);

	return 0;
}

static int set_tty_fd(int fd, struct comm_device *pd, int baud,
                      const tchar_t *devicename) {
	struct termios newt = pd->oldState;
	speed_t speed = B0;
	unsigned i;

	newt.c_cflag = CS8 | CLOCAL | CREAD;

	/* we want line mode (canonical mode) */
	newt.c_lflag |= ICANON;
	/* and no echo */
	newt.c_lflag &= ~ECHO;
	newt.c_iflag |= CLOCAL;

	/*
	 * Baud rate constants (B4800) can be arbitrary
	 * constants (B4800 may not == 4800)
	 */
	for (i = 0; i < ELEMENTS(SPEEDS); i++)
		if (SPEEDS[i].baud == baud) {
			speed = SPEEDS[i].constant;
			break;
		}

	if (cfsetispeed(&newt, speed) < 0) {
		err_push_errno(CFSETISPEED_FN, devicename);
	}

	else if (cfsetospeed(&newt, speed) < 0) {
		err_push_errno(CFSETOSPEED_FN, devicename);
	}

	else if (tcsetattr(fd, TCSAFLUSH, &newt) < 0) {
		err_push_errno(TCSETATTR_FN, devicename);
	}
	else {
		struct termios chkt;

		/*
		 * Serial controllers may not be able to set
		 * speeds until some indeterminate time has
		 * gone by: pl2303 is an example of a remote
		 * controller that does so
		 */
		(void) tcflush(fd, TCIOFLUSH);
		(void) usleep(200000);
		(void) tcflush(fd, TCIOFLUSH);

		memset(&chkt, 0, sizeof(chkt));

		/*
		 * make sure all the serial device settings
		 * have actually taken place
		 */
		if (tcgetattr(fd, &chkt) < 0) {
			err_push_errno(TCGETATTR_FN, devicename);
		}
/*
		else if (memcmp(&newt, &chkt, sizeof(newt)) != 0) {
			err_push(TCSETATTR_FN, devicename,
				 T("Not all attributes set"));
		} */
		else {
			pd->fd = fd;
			pd->running = 1;
			pd->type = comm_tty;
			return 1;
		}
	}

	return 0;
}

#define LINE_LEN (512)

/**
 * Read lines from comm device, passing each line, including
 * terminating LF, to callback function.
 *
 * Returns 0 on error.
 * Returns 1 on EOF.
 * Returns 2 if shut down via comm_stop()
 */
int
comm_loop(struct comm_device *pd,
	  int (*callback)(void *argument, char *line, size_t len),
	  void *argument) {
	char buf[LINE_LEN + 1];
	unsigned i = 0;

	while (pd->running) {
		ssize_t l;
		struct pollfd fds[1];

		fds[0].fd = pd->fd;
		fds[0].events = POLLIN;
		while (poll(fds, 1, 2000) == 0) {
			if (pd->type == comm_tty)
				set_tty_fd(pd->fd, pd, 4800, NULL);
		}
		
		l = read(pd->fd, buf + i, LINE_LEN - i);

		if (l < 0 || l > LINE_LEN - i) {
			err_push_errno(READ_FN, NULL);
			return 0;
		}

		if (l == 0)
			return 1;

		i += l;
		buf[i] = '\0';

		/* now consume as many lines as possible */
		i = str_iter_lines(buf, callback, argument);
	}

	return 2;
}

void
comm_close(struct comm_device *pd) {
	if (pd->type == comm_tty)
		(void) tcsetattr(pd->fd, TCSADRAIN, &pd->oldState);

	(void) close(pd->fd);
}

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
