/*
 * http://www.wayforward.net/avGPS/libav/av_comm.h
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * Low-level OS-specific handling for communications ports.
 * ---
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#ifndef AV_COMM_H
#define AV_COMM_H

#ifdef _WIN32
# define WIN32_LEAN_AND_MEAN
# include <windows.h>

#else
# include <termios.h>

#endif

#include "av_str.h"

#ifdef __cplusplus
extern "C" {
#endif

	enum parity {
		no_parity,
		odd_parity,
		even_parity,
		mark_parity,
		space_parity
	};

	enum stop_bits {
		stop_bits_1_5,
		stop_bits_1,
		stop_bits_2
	};

	struct comm_settings {
		int baud;     /* 4800 */
		int bits;     /* 5, 6, 7, or 8 */
		enum parity parity;
		enum stop_bits stop_bits;
	};

	enum comm_type {
		comm_file,
		comm_socket,
		comm_tty
	};

	struct comm_device {

		enum comm_type type;

		/*
		 * Initially 1, set to 0 if the loop should stop
		 */
		int running;

#ifdef _WIN32
		HANDLE hComm;
		DCB oldState;

#else
		int fd;
		struct termios oldState;

#endif

	};

	/**
	 * Given the device name and some communications settings, try
	 * to open as a comm device.
	 *
	 * Returns 0 on error, with error information pushed onto the av_err
	 * stack.  Returns 1 if successfully open.
	 *
	 * Use comm_is_live() to determine if it's a playback file,
	 * or if it's the real thing.
	 */
	int
	comm_open(const tchar_t *devicename, const struct comm_settings *pc,
		  struct comm_device *pd);

	/**
	 * Read lines from comm device, passing each line, including
	 * terminating LF, to callback function.
	 *
	 * Returns 0 on error.
	 * Returns 1 on EOF.
	 * Returns 2 if shut down via comm_stop()
	 */
	int
	comm_loop(struct comm_device *pd,
		  int (*callback)(void *argument, char *line, size_t len),
		  void *argument);

#define comm_is_live(pd) ((pd)->type!=comm_file)
#define comm_is_running(pd) ((pd)->running)
#define comm_stop(pd) ((pd)->running=0)
#define comm_start(pd) ((pd)->running=1)

	void
	comm_close(struct comm_device *pd);

#ifdef __cplusplus
}
#endif

#endif /* AV_COMM_H */

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
