/**
 * http://www.wayforward.net/avGPS/libav/av_db.h
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * vi:set noexpandtab tabstop=8 shiftwidth=8 softtabstop=8:
 *
 * This is a binary, read-only database of navaids, fixes, obstructions,
 * airports and runways, and routes.
 *
 * The two main indexes into the data are:
 * 1. by name.  Fixes, airports, and runways are indexed by their ident.
 *    a runway's ident is 'ARPT-XX' where 'ARPT' is the four letter ICAO
 *    ident, and XX is the runway name, or heading / 10.  Runways have
 *    duplicate entries based on runway name, so the closest runway 3
 *    can be found.
 * 2. by latitude.  this aids finding all items in the database within
 *    a certain radius of a lat/lon point.
 *
 * The database is maintained by mapping the single database file into
 * memory.
 * ---------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#ifndef AV_DB_H
#define AV_DB_H

#include "av_nav.h"
#include "av_time.h"
#include "av_str.h"

/*
 * The latitude and longitude values in the database are integers:
 * divide by MAX_INT (2147483647) and multiply by either PI or
 * 180, for radians or degrees.
 */

#define IRADIANS(i) ((i) * PI / 2147483647.0)
#define IDEGREES(i) ((i) * 180.0 / 2147483647.0)

#ifdef __cplusplus
extern "C" {
#endif

	enum nav_type {
		nav_t_airport,
		nav_t_runway,
		nav_t_navaid,
		nav_t_fix,
		nav_t_obstruction,
		nav_t_route
	};

	/**
	 * Common header to all following structures
	 */
	struct nav_item {
		unsigned id;
		int ilat, ilon;
	};

#define NAV_TYPE(p) ((enum nav_type)(((const struct nav_item *)(p))->id >> 24))

	struct nav_airport {
		/* struct nav_item */
		unsigned id;
		int ilat, ilon;

		short elevation;
		char ident[4];   /* ex 'KJFK' 'KGLD' 'PHTO' */
		char padding[2];
	};
	/* sizeof(struct nav_airport) == 20 */

	struct nav_runway {
		/* struct nav_item */
		unsigned id;
		int ilat, ilon;

		short tdze, length, width;
		char ident[14];   /* ex 'PHNL-22L' */
	};
	/* sizeof(struct nav_runway) == 32 */

	struct nav_navaid {
		/* struct nav_item */
		unsigned id;
		int ilat, ilon;

		int imagvar;
		unsigned ifreq; /* float freq * 80.0 */

		short elevation;
		char ident[10];
	};
	/* sizeof(struct nav_navaid) == 32 */

	struct nav_fix {
		/* struct nav_item */
		unsigned id;
		int ilat, ilon;

		char ident[8];
	};
	/* sizeof(struct nav_fix) == 20 */

	struct nav_obstruction {
		/* struct nav_item */
		unsigned id;
		int ilat, ilon;

		short msl;
		short agl;
	};
	/* sizeof(struct nav_obstruction) == 16 */

	union nav_union {
		unsigned id;
		struct nav_item item;
		struct nav_airport airport;
		struct nav_obstruction obstruction;
		struct nav_navaid navaid;
		struct nav_fix fix;
		struct nav_runway runway;
	};

	enum nav_route_type {
                nav_route_dp,
                nav_route_star,
                nav_route_victor,
                nav_route_jet,
                nav_route_amber,
                nav_route_green,
                nav_route_blue,
                nav_route_red,
                nav_route_rnav,
                nav_route_iap
	};

	struct nav_route {
		unsigned id;
		char name[16];
		enum nav_route_type type;

		char condition[52];
	};
	/* sizeof(struct nav_route) == 76 */


	/* expand with path_expand() to pass into navdb_init() */
	extern const tchar_t NAVDB_PATH[];

	/**
	 * Loads the nav database.
	 *
	 * Looks at all files in the data/ subdirectory looking for
	 * a file with the correct effective dates.
	 *
	 * Returns 1 on success, otherwise returns 0, and stores the error
	 * code returned on the last database scanned.
	 */
	int
	navdb_init(const tchar_t *const *path, const struct av_datetime *ptd);

	void
	navdb_get_effective(struct av_datetime *dt_start,
			    struct av_datetime *dt_end);

	void
	navdb_add_waypoint(struct nav_position pos);

	/**
	 * Finds a navaid, fix, or airport by name.
	 *
	 * If there are multiple items with the same name, returns
	 * the closest.
	 */
	const struct nav_item *
	navdb_find_ident(const char *name, struct nav_position pos);

	/**
	 * Enumerates all navaids, fixes, airports, etc with a given name.
	 *
	 * Calls _pfn_ with each item that is prefixed by _name_.
	 */
	void
	navdb_enum_prefix(const char *prefix,
			  void (*pfn)(void *pv, const struct nav_item *p),
			  void *pv);

	/**
	 * Finds a navaid, fix, or airport by name.  If there is more than
	 * one entry with the same prefix, find the closest one and return
	 * that.
	 *
	 * Returns NULL if none are found.
	 */
	const struct nav_item *
	navdb_find_nearest_by_prefix(const char *prefix,
				     struct nav_position pos);

	/**
	 * Finds all navaids, airports, runways, obstructions, etc within
	 * a given radius of a lat/long position.
	 */
	void
	navdb_enum_nearest(struct nav_position pos, double radius,
			   void (*pfn)(void *pv, const struct nav_item *p,
				       struct nav_polar *polar),
			   void *pv);

	/**
	 * Given a runway, find the reciprocal runway (same pavement,
	 * different direction).
	 */
	const struct nav_item *
	navdb_other_runway_side(const struct nav_item *p);

	/**
	 * Retrieves text description of nav item.
	 *
	 * If _pnav_ is NULL, sets buf to an empty string and returns 0.
	 * If _pnav_ points to an airport, fix, or navaid, retrieves the
	 * ident and copies it into buf, returning the string length.
	 * If _pnav_ is an obstruction, formats the _msl_ into buf.
	 */
	unsigned
	navdb_get_ident_a(const struct nav_item *pnav,
			  char *buf, unsigned len);

#ifdef UNICODE
	unsigned
	navdb_get_ident_u(const struct nav_item *pnav,
			  wchar_t *buf, unsigned len);

# define navdb_get_ident navdb_get_ident_u
#else
# define navdb_get_ident navdb_get_ident_a
#endif

	struct nav_position
	navdb_position(const struct nav_item *pnav);

	unsigned
	navdb_fix_routes(const struct nav_item *pnav,
			 const struct nav_item *routes[],
			 const struct nav_item *fixes[],
			 unsigned nroutes);

#ifdef __cplusplus
}
#endif

#endif /* AV_DB_H */

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
