/*
 * http://www.wayforward.net/avGPS/libav/av_gps.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * Handle a GPS serial device (or a log file) by parsing it, and
 * signalling via callbacks.
 * --------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include <assert.h>
#include <ctype.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "av_gps.h"
#include "av_comm.h"
#include "av_str.h"

enum {
	gga_tag,
	gga_time,
	gga_lat,
	gga_ns,
	gga_long,
	gga_ew,
	gga_quality,
	gga_satellites,
	gga_hdop,
	gga_alt,
	gga_alt_unit,
	gga_geoid,
	gga_geoid_unit,
	gga_dgps_age,
	gga_dgps_id,

	gga_length
};

enum {
	rmc_tag,
	rmc_time,
	rmc_valid,
	rmc_lat,
	rmc_ns,
	rmc_long,
	rmc_ew,
	rmc_knots,
	rmc_true,
	rmc_date,
	rmc_magvar,
	rmc_unknown,

	rmc_length
};

enum {
	vtg_tag,
	vtg_true,
	vtg_t,
	vtg_magnetic,
	vtg_m,
	vtg_knots,
	vtg_n,
	vtg_kph,
	vtg_k,

	vtg_length
};

enum {
	gsa_tag,
	gsa_op_mode,
	gsa_fix_mode,
	gsa_prns,

	gsa_pdop = 15,
	gsa_hdop,
	gsa_vdop,

	gsa_length
};

static const char LOG_MODE[] = "a";

static const char GGA[] = "GGA";
static const char RMC[] = "RMC";
static const char GSV[] = "GSV";
static const char VTG[] = "VTG";
static const char GSA[] = "GSA";

static const char ZEROES[] = "000000";

static const char GPS_PREFIX[] = "$GP";

static int
gps_line(void *pv, char *line, size_t len);

static void
parse_line(struct gps_device *pd, const struct gps_listener *p,
	   void *pv, char **tokens, unsigned ntokens);

static int
gps_int(const char *str);

static int
gps_int_n(const char *p, unsigned n);

static int
gps_parse_time(const char *time, struct av_time *pt);

static int
gps_parse_datetime(const char *date, const char *time,
		   struct av_datetime *pdt);

static double
gps_parse_latitude(const char *str, const char *ns);

static double
gps_parse_longitude(const char *str, const char *ew);

/**
 * Parse an angle (latitude or longitude) in NMEA-0183 format,
 * as in XXXXYY.YYYY, degrees and decimal minutes.
 *
 * Example:
 * gps_parse_angle("15457.1375") == 154.9523
 */
static double
gps_parse_angle(const char *str);

/**
 * Strips whitespace at end of string.
 *
 * Returns the length of the edited string.
 */
static size_t
strip_r(char *p);

int
gps_open(struct gps_device *pd, const tchar_t *devicename) {
	struct comm_settings settings;

	settings.baud = 4800;
	settings.bits = 8;
	settings.parity = no_parity;
	settings.stop_bits = stop_bits_1;

	memset(pd, 0, sizeof(struct gps_device));
	return comm_open(devicename, &settings, &pd->device);
}

void
gps_close(struct gps_device *pd) {
	comm_close(&pd->device);
}

struct gps_argument {
	struct gps_device *pd;
	const struct gps_listener *pl;
	void *pv;
};

int
gps_loop(struct gps_device *pd, const struct gps_listener *p, void *pv) {
	struct gps_argument a;

	a.pd = pd;
	a.pl = p;
	a.pv = pv;

	pd->quality = 0;
	pd->nsats = 0;

	return comm_loop(&pd->device, gps_line, &a);
}

static int
gps_line(void *pv, char *line, size_t len) {
	size_t l = len;
	struct gps_argument *pa = (struct gps_argument *) pv;
	const struct gps_listener *p = pa->pl;
	struct gps_device *pd = pa->pd;

	char *tokens[32];
	unsigned ntokens;

	pv = pa->pv;

	if (!pd->device.running) {
		return 0;
	}

	if (p->log != NULL)
		(*p->log)(pv, line, len);

	l = strip_r(line);

	/* garbage unless it starts with '$GP' */
	if (memcmp(line, GPS_PREFIX, sizeof(GPS_PREFIX) - 1) != 0) {
		return 0;
	}

	/* checksum is optional */
	if (l > 3 && line[l - 3] == '*') {
		unsigned char testsum = 0;
		unsigned i;

		l -= 3;

		for (i = 1; i < l; i++)
			testsum ^= line[i];

		if (testsum != strtoul(line + l + 1, NULL, 16)) {
			return 0;
		}

		line[l] = '\0';
	}

	ntokens = str_split(line + 3, ',',
			    tokens, ELEMENTS(tokens));

	parse_line(pd, p, pv, tokens, ntokens);

	return 1;
}

static void
parse_line(struct gps_device *pd, const struct gps_listener *p,
	   void *pv, char **tokens, unsigned ntokens) {
	switch(tokens[0][2]) {
	case 'A':
		if (strcmp(tokens[gga_tag], GGA) == 0
		    && ntokens == gga_length) {
			int q;

			if (p->start != NULL)
				(*p->start)(pv);

			q = gps_int(tokens[gga_quality]);

			if (q != pd->quality && p->quality != NULL) {
				struct gps_quality msg;

				msg.type = gps_t_quality;
				msg.quality = (enum gps_fix_quality) q;
				pd->quality = q;
				(*p->quality)(pv, &msg);
			}

			if (p->fix != NULL) {
				struct gps_fix msg;

				msg.type = gps_t_fix;
				msg.quality = (enum gps_fix_quality) q;
				gps_parse_time(tokens[gga_time], &msg.time);
				msg.latitude =
					gps_parse_latitude(tokens[gga_lat],
							   tokens[gga_ns]);
				msg.longitude =
					gps_parse_longitude(tokens[gga_long],
							    tokens[gga_ew]);

				msg.msl_meters = atof(tokens[gga_alt]);
				(*p->fix)(pv, &msg);

				return;
			}
		}
		else if (p->active != NULL
			 && strcmp(tokens[0], GSA) == 0
			 && ntokens == gsa_length) {
			struct gps_active msg;
			unsigned nprns = 0;
			int i;

			msg.type = gps_t_active;
			msg.op_mode = tokens[gsa_op_mode];
			msg.fix_mode = tokens[gsa_fix_mode];

			for (i = gsa_prns; i < gsa_pdop; i++)
				if (tokens[i][0] != '\0')
					msg.prns[nprns++] = atoi(tokens[i]);

			msg.nprns = nprns;

			msg.pdop = atof(tokens[gsa_pdop]);
			msg.hdop = atof(tokens[gsa_hdop]);
			msg.vdop = atof(tokens[gsa_vdop]);

			(*p->active)(pv, &msg);

			return;
		}
		break;

	case 'C':
		if (p->minimum != NULL
		    && strcmp(tokens[rmc_tag], RMC) == 0
		    && ntokens == rmc_length) {

			struct gps_minimum msg;

			gps_parse_datetime(tokens[rmc_date], tokens[rmc_time],
					   &msg.datetime);

			msg.type = gps_t_minimum;
			msg.valid = tokens[rmc_valid][0] == 'A';
			msg.latitude = gps_parse_latitude(tokens[rmc_lat],
							  tokens[rmc_ns]);
			msg.longitude = gps_parse_longitude(tokens[rmc_long],
							    tokens[rmc_ew]);
			msg.knots = atof(tokens[rmc_knots]);

			msg.true_course_valid = (tokens[rmc_true][0] != '\0');
			if (msg.true_course_valid)
				pd->true_course = atof(tokens[rmc_true]);
			msg.true_course = pd->true_course;

			(*p->minimum)(pv, &msg);

			return;
		}
		break;

	case 'V':
		if (p->satellites != NULL
		    && strcmp(tokens[0], GSV) == 0) {

			struct gps_satellite *ps = pd->sats + pd->nsats;
			unsigned i;

			for (i = 4; i < ntokens; i += 4) {
				if (ps >= pd->sats + ELEMENTS(pd->sats))
					break;

				ps->satellite_id =
					gps_int(tokens[i + 0]);
				ps->elevation_degrees =
					gps_int(tokens[i + 1]);
				ps->azimuth_degrees =
					gps_int(tokens[i + 2]);
				ps->snr_dbhz =
					gps_int(tokens[i + 3]);

				ps++;
			}

			pd->nsats = ps - pd->sats;

			if (gps_int(tokens[1]) == gps_int(tokens[2])) {
				if (pd->nsats == gps_int(tokens[3])) {
					struct gps_satellites msg;
					msg.type = gps_t_satellites;
					msg.satellites = pd->sats;
					msg.nsatellites = pd->nsats;

					(*p->satellites)(pv, &msg);
				}

				pd->nsats = 0;
			}

			return;
		}
		break;

	case 'G':
		if (strcmp(tokens[vtg_tag], VTG) == 0
		    && ntokens == vtg_length) {

			if (p->course != NULL) {
				struct gps_course msg;
				char *ptrue = tokens[vtg_true];
				char *pknots = tokens[vtg_knots];

				msg.type = gps_t_course;

				msg.true_course_valid = (ptrue[0] != '\0');
				if (msg.true_course_valid)
					msg.true_course = atof(ptrue);

				msg.knots_valid = (pknots[0] != '\0');

				if (msg.knots_valid)
					msg.knots = atof(pknots);

				(*p->course)(pv, &msg);
			}
			else if (p->unhandled != NULL)
				(*p->unhandled)(pv, tokens, ntokens);

			if (p->end != NULL)
				(*p->end)(pv);

			return;
		}
		break;

	}

	if (p->unhandled != NULL)
		(*p->unhandled)(pv, tokens, ntokens);
}

static int
gps_int(const char *str) {
	char *p;
	long l = strtol(str, &p, 10);

	return p == str ? -1 : (int) l;
}

static int
gps_int_n(const char *p, unsigned n) {
	const char *pe = p + n;
	int result = 0;

	while (p != pe) {
		char c = *p++;
		if ('0' <= c && c <= '9')
			result = result * 10 + c - '0';
		else
			break;
	}

	return result;
}

static int
gps_parse_time(const char *time, struct av_time *pt) {
	unsigned c;

	c = strlen(time);
	if (c < 6 || 13 < c) return 0;

	pt->hour = gps_int_n(time + 0, 2);
	pt->min = gps_int_n(time + 2, 2);
	pt->sec = gps_int_n(time + 4, 2);

	if (c > 7 || time[6] == '.') {
		char usecs[14];

		strncpy(usecs, time + 7, 6);
		strcat(usecs, ZEROES);
		pt->usecs = gps_int_n(usecs, 6);
	}
	else
		pt->usecs = 0;

	return 1;
}

static int
gps_parse_datetime(const char *date, const char *time,
		   struct av_datetime *pdt) {
	if (!gps_parse_time(time, &pdt->time))
		return 1;

	pdt->date.day   = gps_int_n(date + 0, 2);
	pdt->date.month = gps_int_n(date + 2, 2);
	pdt->date.year  = gps_int_n(date + 4, 2) + 2000;

	return 1;
}

static double
gps_parse_latitude(const char *str, const char *ns) {
	static const char S[] = "S";

	double lat = gps_parse_angle(str);

	return strcmp(ns, S) == 0 ? -lat : lat;
}

static double
gps_parse_longitude(const char *str, const char *ew) {
	static const char W[] = "W";

	double lon = gps_parse_angle(str);

	return strcmp(ew, W) == 0 ? -lon : lon;
}

/**
 * Parse an angle (latitude or longitude) in NMEA-0183 format,
 * as in XXXXYY.YYYY, degrees and decimal minutes.
 *
 * Example:
 * gps_parse_angle("15457.1375") == 154.9523
 */
static double
gps_parse_angle(const char *str) {
	char *p;
	unsigned u = strtoul(str, &p, 10);
	double d;

	if (*p != '.' || p < str + 3) return 0.0;

	d = atof(p - 2);

	return u / 100 + d / 60.0;
}

/**
 * Strips whitespace at end of string.
 *
 * Returns the length of the edited string.
 */
static size_t
strip_r(char *p) {
	size_t l = strlen(p);

	while (l > 0) {
		l--;

		if (!isspace(p[l]))
			return l + 1;

		p[l] = '\0';
	}

	return l;
}

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
