/**
 * http://www.wayforward.net/avGPS/libav/av_str.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
# define WIN32_LEAN_AND_MEAN
# include <windows.h>
#endif

#include "av_str.h"

static size_t
expanded_string_length(const tchar_t *p, const tchar_t *e);

static const tchar_t *
copy_until(const tchar_t *p, const tchar_t *e, tchar_t terminator,
	   tchar_t *buf, size_t buflen);

static const tchar_t *
copy_ident(const tchar_t *p, const tchar_t *e,
	   tchar_t *buf, size_t buflen);

static const tchar_t HOME[] = T("HOME");

/**
 * Separate a string into component parts by replacing all
 * ':' or separator characters with nulls.  Returns the
 * number of strings.
 *
 * Examples:
 *    original(p)   delim(ch)       result(p)        return
 *    -----------   ---------       ---------        ------
 *    foo               :           foo                 1
 *    foo:bar           :           foo\0bar            2
 *
 *    $GPRMC,FOO,,BAR   ,           $GPRMC\0FOO\0\0BAR  4
 *
 * If the _tokens_ array is non-null, as many split strings
 * as possible are stored in it.
 *
 * Example:
 *    original(p)   delim(ch)       result(p)        return
 *    -----------   ---------       ---------        ------
 *    a,b,c,d,e,f       ,           a\0b\0c\0d\0e\0f    6
 *
 *    if _ntokens_ is 4:
 *    tokens[0] == "a"
 *    tokens[1] == "b"
 *    tokens[2] == "c"
 *    tokens[3] == "d\0e\0f"
 */
unsigned
tstr_split(tchar_t *p, tchar_t ch, tchar_t **tokens, unsigned ntokens) {
	unsigned c = 0;

	if (c < ntokens) tokens[c] = p;
	c++;

	while (*p) {
		if (*p == ch) {
			*p++ = T('\0');
			if (c < ntokens) tokens[c] = p;
			c++;
		}
		else
			p++;
	}

	return c;
}

#ifdef UNICODE
unsigned
str_split(char *p, char ch, char **tokens, unsigned ntokens) {
	unsigned c = 0;

	if (c < ntokens) tokens[c] = p;
	c++;

	while (*p) {
		if (*p == ch) {
			*p++ = '\0';
			if (c < ntokens) tokens[c] = p;
			c++;
		}
		else
			p++;
	}

	return c;
}
#endif

#ifndef _WIN32
int
av_getenv(const tchar_t *name, tchar_t *buf, size_t len) {
	char *env = getenv(name);
	if (env != NULL) {
		size_t l = strlen(env);
		if (l >= len)
			l = len - 1;
		memcpy(buf, env, l * sizeof(tchar_t));
		buf[l] = T('\0');
		return l;
	}
	else {
		buf[0] = '\0';
		return 0;
	}
}
#endif

tchar_t *
str_expand(const tchar_t *p, const tchar_t *e) {
	/* two passes: one to size the string, one to copy the string */
	size_t l = expanded_string_length(p, e);
	tchar_t *result = malloc((l + 1) * sizeof(tchar_t));
	tchar_t *r_p = result, *r_e = result + l;
	tchar_t buf[512];

	if (result == NULL)
		return result;

	if (p != e
	    && p[0] == T('~')
	    && ((p + 2 <= e && p[1] == DIR_SEP)
		|| p + 1 == e
		|| p[1] == T('\0'))) {
		av_getenv(HOME, buf, ELEMENTS(buf));
		r_p = str_append(buf, r_p, r_e);
		p++;
	}

	while (p != e) {
		tchar_t ch = *p++;

		if (ch == T('\0'))
			break;

		if (ch == T('$')) {
			tchar_t var[128];

			if (p == e)
				break;

			if (*p == T('('))
				p = copy_until(p + 1, e, T(')'),
					       var, ELEMENTS(var));
			else if (*p == T('{'))
				p = copy_until(p + 1, e, T('}'),
					       var, ELEMENTS(var));
			else
				p = copy_ident(p, e, var, ELEMENTS(var));

			av_getenv(var, buf, ELEMENTS(buf));
			r_p = str_append(buf, r_p, r_e);
		}
		else if (r_p != r_e)
			*r_p++ = ch;
	}

	*r_p = T('\0');

	return result;
}

static size_t
expanded_string_length(const tchar_t *p, const tchar_t *e) {
	size_t result = 0;
	tchar_t buf[512];

	if (p != e) {
		if (p[0] == T('~') && ((p + 2 <= e && p[1] == DIR_SEP)
				       || p + 1 == e
				       || p[1] == T('\0'))) {
			result += av_getenv(HOME, buf, ELEMENTS(buf));
			p++;
		}

		while (p != e) {
			tchar_t ch = *p++;

			if (ch == '\0')
				break;

			if (ch == '$') {
				tchar_t var[128];

				if (p == e)
					break;

				if (*p == '(')
					p = copy_until(p + 1, e, ')',
						       var, ELEMENTS(var));
				else if (*p == '{')
					p = copy_until(p + 1, e, '}',
						       var, ELEMENTS(var));
				else
					p = copy_ident(p, e, var,
						       ELEMENTS(var));

				result += av_getenv(var, buf,
						    ELEMENTS(buf));
			}
			else
				result++;
		}
	}

	return result;
}

static const tchar_t *
copy_until(const tchar_t *p, const tchar_t *e, tchar_t terminator,
	   tchar_t *buf, size_t buflen) {
	tchar_t *bufe = buf + buflen - 1;

	assert(buflen > 0);

	while (p != e) {
		tchar_t ch = *p++;

		if (ch == terminator)
			break;

		if (buf != bufe)
			*buf++ = ch;
	}

	*buf = T('\0');

	return p;
}

static const tchar_t *
copy_ident(const tchar_t *p, const tchar_t *e,
	   tchar_t *buf, size_t buflen) {
	tchar_t *bufe = buf + buflen - 1;

	assert(buflen > 0);

	while (p != e) {
		tchar_t ch = *p;

		if (!isalnum(ch) && ch != '_')
			break;

		if (buf != bufe)
			*buf++ = ch;

		p++;
	}

	*buf = '\0';

	return p;
}

/**
 * Append the _src_ string to the buffer marked by
 * _p_ (beginning) to _e_ (end).
 */
tchar_t *
str_append(const tchar_t *src, tchar_t *p, tchar_t *e) {
	if (src != NULL)
		while (*src && p != e)
			*p++ = *src++;

	return p;
}

unsigned
str_count_chars(const tchar_t *str, tchar_t ch) {
	unsigned result = 0;

	while (*str) {
		if (*str == ch)
			result++;
		str++;
	}

	return result;
}

tchar_t *
tstr_save(const tchar_t *str) {
	tchar_t *p;
	if (str != NULL) {
		size_t l = tstrlen(str) + 1;

		p = malloc(l * sizeof(tchar_t));

		if (p != NULL)
			memcpy(p, str, l * sizeof(tchar_t));
	}
	else
		p = NULL;

	return p;
}

int
str_ends_with(const tchar_t *p, const tchar_t *e, const tchar_t *end) {
	size_t l = tstrlen(end);

	if (e - l < p) return 0;

	assert(e - l >= p);

	return tstrcmp(e - l, end) == 0;
}

size_t
str_iter_lines(char *buffer,
	       int (*callback)(void *argument, char *line, size_t len),
	       void *argument) {
	char *pline = buffer, *p;
	size_t i;

	while ((p = strchr(pline, '\n')) != NULL) {
		char ch = *++p;

		*p = '\0';

		(*callback)(argument, pline, p - pline);

		*p = ch;
		pline = p;
	}

	if (pline == buffer)
		return strlen(buffer);

	for (i = 0; pline[i]; i++)
		buffer[i] = pline[i];
	buffer[i] = '\0';
	return i;
}

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
