/**
 * http://www.wayforward.net/avGPS/libav/av_str.h
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#ifndef AV_STR_H
#define AV_STR_H

#include <stddef.h>

#if defined(UNICODE)
# include <wchar.h>
# define tchar_t wchar_t
# define T(str) L##str
# define tstrlen wcslen
# define tstrcpy wcscpy
# define tstrncpy wcsncpy
# define tstrcmp wcscmp
# define tstrcat wcscat
# define tstrstr wcsstr
# define tstrchr wcschr
# define tstrrchr wcsrchr
# define tstrtoul wcstoul
# define tstrtol wcstol
# define stprintf swprintf

#else
# define tchar_t char
# define T(str) str
# define tstrlen strlen
# define tstrcpy strcpy
# define tstrncpy strncpy
# define tstrcmp strcmp
# define tstrcat strcat
# define tstrstr strstr
# define tstrchr strchr
# define tstrrchr strrchr
# define tstrtoul strtoul
# define tstrtol strtol
# define stprintf sprintf

#endif

#ifdef _WIN32
# define PATH_SEP T(';')
# define DIR_SEP T('\\')

#else
# define PATH_SEP T(':')
# define DIR_SEP  T('/')

#endif

#define ELEMENTS(array) (sizeof(array)/sizeof((array)[0]))

#ifdef __cplusplus
extern "C" {
#endif

	/**
	 * Malloc'd copy of string _str_.  Returns NULL if out
	 * of memory.
	 */
	tchar_t *
	tstr_save(const tchar_t *str);

	/**
	 * Separate a string into component parts by replacing all
	 * ':' or separator characters with nulls.  Returns the
	 * number of strings.
	 *
	 * Examples:
	 *    original(p)   delim(ch)       result(p)        return
	 *    -----------   ---------       ---------        ------
	 *    foo               :           foo                 1
	 *    foo:bar           :           foo\0bar            2
	 *
	 *    $GPRMC,FOO,,BAR   ,           $GPRMC\0FOO\0\0BAR  4
	 *
	 * If the _tokens_ array is non-null, as many split strings
	 * as possible are stored in it.
	 *
	 * Example:
	 *    original(p)   delim(ch)       result(p)        return
	 *    -----------   ---------       ---------        ------
	 *    a,b,c,d,e,f       ,           a\0b\0c\0d\0e\0f    6
	 *
	 *    if _ntokens_ is 4:
	 *    tokens[0] == "a"
	 *    tokens[1] == "b"
	 *    tokens[2] == "c"
	 *    tokens[3] == "d\0e\0f"
	 */
	unsigned
	tstr_split(tchar_t *p, tchar_t ch, tchar_t **tokens, unsigned ntokens);

#ifdef UNICODE
	unsigned
	str_split(char *p, char ch, char **tokens, unsigned ntokens);
#else
# define str_split tstr_split
#endif

	/**
	 * Expand shell escape sequences and variables.
	 *
	 * Returns a newly-malloc'd string to hold the expansion.
	 * Call free() to release the string.
	 *
	 * Examples:
	 *   str_expand("~/foo", NULL) -> "/Users/terry/foo"
	 *   str_expand("$(JAVA_HOME)", NULL) -> "/usr/local/jdk-1.4/"
	 */
	tchar_t *
	str_expand(const tchar_t *p, const tchar_t *e);

	/**
	 * Append the _src_ string to the buffer marked by
	 * _p_ (beginning) to _e_ (end).
	 */
	tchar_t *
	str_append(const tchar_t *src, tchar_t *p, tchar_t *e);

	/**
	 * Count the number of times _ch_ appears in _str_.
	 */
	unsigned
	str_count_chars(const tchar_t *str, tchar_t ch);

	int
	str_ends_with(const tchar_t *p, const tchar_t *e, const tchar_t *end);

	size_t
	str_iter_lines(char *buffer,
		       int (*callback)(void *argument, char *line, size_t len),
		       void *argument);

#ifdef UNICODE
	wchar_t *
	utf8_to_unicode(wchar_t *dst, const char *src, size_t cdst);
#endif

	/**
	 * Binary search for _key_ in array _base_.  The _base_ array consists
	 * of _count_ items, each of _size_ bytes, and must be sorted.
	 *
	 * _cmp_ is a function which should return:
	 * - a negative value if _p1_ should be ordered before _p2_,
	 * - 0 if _p1_ is the same as _p2_, or
	 * - a positve value if _p1_ is after _p2_.
	 *
	 * If no item can be found (_cmp_ never returns 0) then NULL is
	 * returned. Otherwise the pointer into the _base_ array that
	 * compares == 0 is returned.
	 */
	const void *
	av_bsearch(const void *key, const void *base, size_t count, size_t size,
		   int (*cmp)(const void *key, const void *item));

/*
 * Windows/CE/Mobile/whatever it's called has no environment.
 * So we just fake it.
 */
#ifdef _WIN32_WCE
# define av_getenv(name, buf, len) (0)

#else
# ifdef _WIN32
#  define av_getenv GetEnvironmentVariable

# else
	int
	av_getenv(const tchar_t *name, tchar_t *buf, size_t len);

# endif
#endif

#ifdef _WIN32
# define tsprintf wsprintf
#else
# ifdef UNICODE
#  define tsprintf swprintf
# else
#  define tsprintf sprintf
# endif
#endif

#ifdef __cplusplus
}
#endif

#endif /* AV_STR_H */

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
