#include <Python.h>

#include "av_db.h"
#include "av_time.h"
#include "av_err.h"
#include "av_mag.h"

static PyObject *
avdata_init(PyObject *self, PyObject *args);

static PyObject *
avdata_nearest_by_ident(PyObject *self, PyObject *args);

static PyObject *
avdata_nearest_by_prefix(PyObject *self, PyObject *args);

static PyObject *
avdata_magnetic_variation(PyObject *self, PyObject *args);

static PyObject *
avdata_routes(PyObject *self, PyObject *args);

static PyObject *
avdata_nearest(PyObject *self, PyObject *args);

static PyObject *
py_nav_item(const struct nav_item *p);

static PyObject *
py_airport(const struct nav_item *p);

static PyObject *
py_runway(const struct nav_item *p);

static PyObject *
py_navaid(const struct nav_item *p);

static PyObject *
py_fix(const struct nav_item *p);

static PyObject *
py_obstruction(const struct nav_item *p);

static PyObject *
py_route(const struct nav_item *p);

static void
enum_nearest(void *pv, const struct nav_item *p, struct nav_polar *polar);

static PyMethodDef avdata_methods[] = {
    {"init",               avdata_init,               METH_VARARGS,
     "init(path) -> (start date tuple, end date tuple)\n\n"
     "Initialize the database, loading data given in a directory or a list\n"
     "of directories"},

    {"nearest_by_ident",   avdata_nearest_by_ident,   METH_VARARGS,
     "nearest_by_ident(ident, (latitude, longitude)) -> NavItem\n\n"
     "Returns the nearest NavItem with the exact name.  _latitude_ and\n"
     "_longitude_ are in radians."},

    {"nearest_by_prefix",  avdata_nearest_by_prefix,  METH_VARARGS,
     "nearest_by_prefix(str, (latitude, longitude) -> NavItem\n\n"
     "Returns the nearest NavItem whose ident starts with _str_.  _latitude_\n"
     "and _longitude_ are in radians."},

    {"magnetic_variation", avdata_magnetic_variation, METH_VARARGS,
     "magnetic_variation(date, (latitude radians, longitude radians), meters_msl)"
     "\n   -> radians\n\n"
     "Calculates the magnetic variation at a point.  East is least: negative\n"
     "values imply east variation."},
    {"routes",             avdata_routes,             METH_VARARGS},
    {"nearest",            avdata_nearest,            METH_VARARGS,
     "nearest((latitude, longitude), nm) -> [NavItem, Obstruction list]\n\n"
     "Returns a list of obstructions and navitems within _nm_ nautical miles\n"
     "of _latitude_ and _longitude_, in radians."},

    {NULL, NULL, NULL, NULL}
};

static PyObject *avdata_module;
static PyObject *avdata_cache;
static PyObject *avdata_Airport;
static PyObject *avdata_Runway;
static PyObject *avdata_Navaid;
static PyObject *avdata_Fix;
static PyObject *avdata_Obstruction;
static PyObject *avdata_Route;

void
init_avdata(void) {
    Py_InitModule("_avdata", avdata_methods);
}

#define AVDATA_MAX_PATH (64)

static PyObject *
avdata_init(PyObject *self, PyObject *args) {
    static const char avdata_init_args[] = "O:init";
    static const char avdata_init_error[] = "init() argument 1 must be a "
                                            "string or a list of strings";
    static const char avdata_init_format[] = "((iiiiiii)(iiiiiii))";

    PyObject *list;
    const char *path[AVDATA_MAX_PATH];
    unsigned npath = 0;
    struct av_datetime start, end;

    if (!PyArg_ParseTuple(args, avdata_init_args, &list))
        return NULL;

    if (PyList_Check(list)) {
        Py_ssize_t i, n = PyList_Size(list);

        if (n >= AVDATA_MAX_PATH) n = AVDATA_MAX_PATH - 1;

        for (i = 0; i < n; i++) {
            PyObject *o = PyList_GET_ITEM(list, i);
            if (!PyString_Check(o)) {
                PyErr_SetString(PyExc_TypeError, avdata_init_error);
                return NULL;
            }

            path[npath++] = PyString_AsString(o);
         }
    }
    else if (PyTuple_Check(list)) {
        Py_ssize_t i, n = PyTuple_Size(list);

        if (n >= AVDATA_MAX_PATH) n = AVDATA_MAX_PATH - 1;

        for (i = 0; i < n; i++) {
            PyObject *o = PyTuple_GET_ITEM(list, i);
            if (!PyString_Check(o)) {
                PyErr_SetString(PyExc_TypeError, avdata_init_error);
                return NULL;
            }

            path[npath++] = PyString_AsString(o);
        }
    }
    else if (PyString_Check(list)) {
        path[npath++] = PyString_AsString(list);
    }
    else {
        PyErr_SetString(PyExc_TypeError, avdata_init_error);
        return NULL;
    }

    err_init();

    path[npath++] = NULL;
    if (!navdb_init(path, NULL)) {
        PyErr_SetString(PyExc_RuntimeError, err_text());
        return NULL;
    }

    navdb_get_effective(&start, &end);

    avdata_module = PyImport_ImportModule("avdata");
    avdata_cache = PyObject_GetAttrString(avdata_module, "cache");
    avdata_Airport = PyObject_GetAttrString(avdata_module, "Airport");
    avdata_Runway = PyObject_GetAttrString(avdata_module, "Runway");
    avdata_Navaid = PyObject_GetAttrString(avdata_module, "Navaid");
    avdata_Fix = PyObject_GetAttrString(avdata_module, "Fix");
    avdata_Obstruction = PyObject_GetAttrString(avdata_module, "Obstruction");
    avdata_Route = PyObject_GetAttrString(avdata_module, "Route");

    return Py_BuildValue(avdata_init_format,
                         start.date.year, start.date.month, start.date.day,
                         start.time.hour, start.time.min, start.time.sec,
                         start.time.usecs,
                         end.date.year, end.date.month, end.date.day,
                         end.time.hour, end.time.min, end.time.sec,
                         end.time.usecs);
}


static PyObject *
py_nav_item(const struct nav_item *p) {
    if (p == NULL) {
        Py_INCREF(Py_None);
        return Py_None;
    }
    else {
        switch (NAV_TYPE(p)) {
        case nav_t_airport:
            return py_airport(p);

        case nav_t_runway:
            return py_runway(p);

        case nav_t_navaid:
            return py_navaid(p);

        case nav_t_fix:
            return py_fix(p);

        case nav_t_obstruction:
            return py_obstruction(p);

        case nav_t_route:
            return py_route(p);

        default:
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
}

static PyObject *
py_airport(const struct nav_item *p) {
    static const char py_airport_format[] = "(Ois(ff)i)";

    const struct nav_airport *pa = (const struct nav_airport *) p;

    char buffer[5];

    navdb_get_ident(p, buffer, sizeof(buffer));

    return PyEval_CallObject(avdata_cache,
                             Py_BuildValue(py_airport_format, avdata_Airport,
                                           pa->id, buffer,
                                           IRADIANS(pa->ilat),
                                           IRADIANS(pa->ilon),
                                           pa->elevation));
}

static PyObject *
py_runway(const struct nav_item *p) {
    static const char py_runway_format[] = "(Ois(ff)iii)";

    const struct nav_runway *pr = (const struct nav_runway *) p;

    char buffer[16];

    navdb_get_ident(p, buffer, sizeof(buffer));

    return PyEval_CallObject(avdata_cache,
                             Py_BuildValue(py_runway_format, avdata_Runway,
                                           pr->id, buffer,
                                           IRADIANS(pr->ilat),
                                           IRADIANS(pr->ilon),
                                           pr->tdze, pr->length, pr->width));
}

static PyObject *
py_navaid(const struct nav_item *p) {
    static const char py_navaid_format[] = "(Ois(ff)isf)";
    static const char freq_format[] = "%.1f";

    const struct nav_navaid *pn = (const struct nav_navaid *) p;

    char buffer[8], freq[12];

    navdb_get_ident(p, buffer, sizeof(buffer));
    sprintf(freq, freq_format, pn->ifreq / 80.0);

    return PyEval_CallObject(avdata_cache,
                             Py_BuildValue(py_navaid_format, avdata_Navaid,
                                           pn->id, buffer,
                                           IRADIANS(pn->ilat),
                                           IRADIANS(pn->ilon),
                                           pn->elevation, freq,
                                           IRADIANS(pn->imagvar)));
}

static PyObject *
py_fix(const struct nav_item *p) {
    static const char py_fix_format[] = "(Ois(ff))";

    const struct nav_fix *pf = (const struct nav_fix *) p;

    char buffer[8];

    navdb_get_ident(p, buffer, sizeof(buffer));

    return PyEval_CallObject(avdata_cache,
                             Py_BuildValue(py_fix_format, avdata_Fix, pf->id,
                                           buffer,
                                           IRADIANS(pf->ilat),
                                           IRADIANS(pf->ilon)));
}

static PyObject *
py_obstruction(const struct nav_item *p) {
    static const char py_obstruction_format[] = "(Oi(ff)ii)";

    const struct nav_obstruction *po = (const struct nav_obstruction *) p;

    return PyEval_CallObject(avdata_cache,
                             Py_BuildValue(py_obstruction_format,
                                           avdata_Obstruction, po->id,
                                           IRADIANS(po->ilat),
                                           IRADIANS(po->ilon),
                                           po->msl, po->agl));
}

static PyObject *
py_route(const struct nav_item *p) {
    static const char py_route_format[] = "(Oisi)";

    const struct nav_route *pr = (const struct nav_route *) p;

    char buffer[44];

    navdb_get_ident(p, buffer, sizeof(buffer));
    return PyEval_CallObject(avdata_cache,
                             Py_BuildValue(py_route_format,
                                           avdata_Route, pr->id, buffer,
                                           pr->type));
}

static PyObject *
avdata_nearest_by_ident(PyObject *self, PyObject *args) {
    static const char avdata_nearest_by_ident_args[] = "s(dd):nearest_by_ident";

    const char *ident;
    struct nav_position pos;

    if (!PyArg_ParseTuple(args, avdata_nearest_by_ident_args,
                          &ident, &pos.latitude, &pos.longitude))
        return NULL;

    return py_nav_item(navdb_find_ident(ident, pos));
}

static PyObject *
avdata_nearest_by_prefix(PyObject *self, PyObject *args) {
    static const char avdata_nearest_by_prefix_args[] = "s(dd):nearest_by_prefix";

    const char *prefix;
    struct nav_position pos;

    if (!PyArg_ParseTuple(args, avdata_nearest_by_prefix_args,
                          &prefix, &pos.latitude, &pos.longitude))
        return NULL;

    return py_nav_item(navdb_find_nearest_by_prefix(prefix, pos));
}

static PyObject *
avdata_magnetic_variation(PyObject *self, PyObject *args) {
    static const char magvar_format[] = "f";
    static const char avdata_magnetic_variation_args[] = "d(dd)d:magnetic_variation";

    static double dt, lat, lon, elev;

    if (!PyArg_ParseTuple(args, avdata_magnetic_variation_args,
                          &dt, &lat, &lon, &elev))
        return NULL;

    return Py_BuildValue(magvar_format, magnetic_variation(dt, lat, lon, elev));
}

static PyObject *
avdata_routes(PyObject *self, PyObject *args) {
    struct nav_item cheat;
    static const char avdata_routes_args[] = "i:routes";
    static const char item_format[] = "(OO)";

    const struct nav_item *routes[1024];
    const struct nav_item *fixes[1024];
    PyObject *array;

    unsigned i, k;

    if (!PyArg_ParseTuple(args, avdata_routes_args, &cheat.id))
        return NULL;

    i = navdb_fix_routes(&cheat, routes, fixes, ELEMENTS(routes));

    if (i > 1024) i = 1024;
    array = PyList_New(i);

    for (k = 0; k < i; k++)
        PyList_SET_ITEM(array, k,
                        Py_BuildValue(item_format, py_nav_item(routes[k]),
                                      py_nav_item(fixes[k])));

    return array;
}

struct nearest_context {
    unsigned count;
    PyObject *items[1024];
};

static PyObject *
avdata_nearest(PyObject *self, PyObject *args) {
    static const char avdata_nearest_args[] = "(dd)d:nearest";

    struct nav_position pos;
    double radius;
    struct nearest_context c;
    PyObject *array;
    unsigned i;

    c.count = 0;

    if (!PyArg_ParseTuple(args, avdata_nearest_args,
                          &pos.latitude, &pos.longitude, &radius))
        return NULL;

    navdb_enum_nearest(pos, radius, enum_nearest, &c);

    array = PyList_New(c.count);

    for (i = 0; i < c.count; i++)
        PyList_SET_ITEM(array, i, c.items[i]);

    return array;
}

static void
enum_nearest(void *pv, const struct nav_item *p, struct nav_polar *polar) {
    static const char AZIMUTH[] = "azimuth";
    static const char DISTANCE[] = "distance";
    static const char DOUBLE[] = "d";

    struct nearest_context *pc = (struct nearest_context *) pv;

    if (pc->count < 1024) {
        PyObject *po = py_nav_item(p);
        PyObject_SetAttrString(po, AZIMUTH,
                               Py_BuildValue(DOUBLE, polar->angle));
        PyObject_SetAttrString(po, DISTANCE,
                               Py_BuildValue(DOUBLE, polar->distance));

        pc->items[pc->count++] = po;
    }
}
