#!/usr/bin/env python3

import datetime

class ParseError(Exception):
    pass

def parse_record(o, line, records):
    for start, length, func, attrib in records:
        substr = line[start:start+length]
        try:
            value = func(substr)
        except:
            raise ParseError(attrib, substr)
        else:
            setattr(o, attrib, value)

    return o

def parse_ljust_an(str):
    """Handle a fadc.faa.gov data file LJUST AN (left-justified alphanumeric)
    field.

    >>> parse_ljust_an("test     ")
    'test'

    >>> parse_ljust_an(" test    ")
    ' test'
    """
    return str.rstrip()

def parse_opt_int(str):
    str = str.strip()
    if str:
        return int(str)
    else:
        return None

def parse_opt_float(str):
    str = str.strip()
    if str:
        return float(str)
    else:
        return None

def parse_date(str):
    """Parse a date in MM/DD/YYYY format.

    >>> parse_date("07/12/1967")
    datetime.date(1967, 7, 12)
    """
    m, d, y = str.split('/', 2)
    return datetime.date(int(y), int(m), int(d))

def parse_none(str):
    return None

def parse_magvar(str):
    sign = str[-1]
    if sign == 'W':
        return float(str[:-1])
    elif sign == 'E':
        return -float(str[:-1])
    else:
        assert str.strip() == ''
        return None

def parse_latitude_seconds(str):
    """Parse a latitude in seconds format.

    >>> parse_latitude_seconds("70992.0N")
    19.72

    >>> parse_latitude_seconds(" 70992.0S ")
    -19.72
    """
    return _parse_seconds(str.strip(), 'N', 'S')

def parse_longitude_seconds(str):
    """Parse a longitude in seconds format.

    >>> round(parse_longitude_seconds(" 558174.0W "), 3)
    -155.048
    """
    return _parse_seconds(str.strip(), 'E', 'W')

def _parse_seconds(str, pos, neg):
    if str[-1] == pos:
        return float(str[:-1]) / 3600.0
    else:
        assert str[-1] == neg
        return float(str[:-1]) / -3600.0

def parse_opt_lat_seconds(str):
    return _parse_opt_seconds(str, 'N', 'S')

def parse_opt_lon_seconds(str):
    return _parse_opt_seconds(str, 'E', 'W')

def _parse_opt_seconds(str, pos, neg):
    str = str.strip()
    if not str:
        return None
    else:
        return _parse_seconds(str, pos, neg)

def parse_latitude_hms(str, sep='-'):
    """Parse latitude in HH-MM-SS[NS] format.

    >>> parse_latitude_hms(" 19-43-12N ")
    19.72
    """
    return _parse_hms(str.strip(), 'N', 'S', sep)

def parse_longitude_hms(str, sep='-'):
    """Parse longitude in HH-MM-SS[EW] format.

    >>> round(parse_longitude_hms("155-02-54W"), 3)
    -155.048
    """
    return _parse_hms(str.strip(), 'E', 'W', sep)

def parse_opt_lat_hms(str):
    return _parse_opt_hms(str, 'N', 'S')

def parse_opt_lon_hms(str):
    return _parse_opt_hms(str, 'E', 'W')

def _parse_opt_hms(str, pos, neg):
    str = str.strip()
    if not str:
        return None
    else:
        return _parse_hms(str, pos, neg, '-')

def _parse_hms(str, pos, neg, sep):
    sign = str[-1]
    h, m, s = str[:-1].split(sep, 3)

    result = float(h) + float(m) / 60.0 + float(s) / 3600.0

    if str[-1] == pos:
        return result
    else:
        assert str[-1] == neg
        return -result

def _test():
    import doctest
    return doctest.testmod()

if __name__ == '__main__':
    _test()
