#!/usr/bin/env python3
#
#
# Copyright (c) 2006, 2010 Wayforward LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA

"""Flight Planning Software.
Helps for cross-country flights.
"""

__author__ = "Terence Way"
__email__ = "terry@wayforward.net"
__version__ = "0.1: November 12, 2010"

# A brief note about rounding and accuracy: we round to the most
# conservative estimate for fuel consumption.  For example, when
# calculating speeds, we round DOWN to the nearest knot, when
# calculating distances, we round UP to the nearest 1/10 nautical mile.

from math import acos, asin, atan2, cos, degrees, floor, fmod, pi, radians, \
                 sin, sqrt

def cartesian(a, d):
    """Return (x, y) cartesian coordinates given polar coordinates
    (angle, distance).
    """
    return sin(a) * d, cos(a) * d

def polar(x, y):
    return atan2(x, y), sqrt(x*x + y*y)

def wind1(course_gs, heading_tas):
    """Given an airplane's GPS-derived course and ground-speed,
    calculate the wind necessary to show a particular heading and
    true airspeed.
    """
    course, gs = course_gs
    heading, tas = heading_tas
    x1, y1 = cartesian(course, gs)
    x2, y2 = cartesian(heading, tas)
    return polar(x2 - x1, y2 - y1)

def wind(course_tas, wd_ws):
    """Given an airplane's vector (course, tas) correct for wind
    given the wind's vector (wd: direction, ws: speed).

    Return a tuple (hd: heading, gs: ground speed).

    Speeds can be in any units: knots, mph, furlongs per fortnight, but
    they must be consistent.  course and wd must be in degrees 0.0..360.0

    Examples:

    If we're in a plane with a TAS (true airspeed) of 100 knots heading
    north (0 degrees), and the wind is at 000@20 (from the north at 20
    knots):
    >>> wind( (0, 100), (0, 20) )
    (0.0, 80.0)

    Our ground speed is 80 knots, with no course correction.

    What if the wind was from the east, at 090@20?
    >>> a, d = wind( (0, 100), (radians(90), 20) )
    >>> round(degrees(a))
    12
    >>> floor(d)
    97

    More examples, from an introduction to the E6B:
    >>> a, d = wind( (radians(205), 160), (radians(240), 25) )
    >>> round(degrees(a))
    210
    >>> floor(d)
    138

    >>> a, d = wind( (radians(137), 88), (radians(240), 16) )
    >>> round(degrees(a))
    147
    >>> floor(d)
    90
    """
    course, tas = course_tas

    # short circuit if TAS is 0
    if tas == 0.0:
        return wd_ws

    else:
        wd, ws = wd_ws
        # rwind is relative wind, in radians
        rwind = wd - course

        tx = ws * cos(rwind)
        ty = ws * sin(rwind)

        wca = asin(ty / tas)

        return (course + wca) % (2 * pi), tas * cos(wca) - tx

def great_circle(pos1, pos2):
    """Calculate distance (in nautical miles) and course (in radians)
    between two points on the Earth, using a great circle path.

    Returns (nm, radians).

    For example, what is the great circle path from LAX (33 degrees,
    57 minutes North, 118 degrees, 24 minutes West) to JFK (40
    degrees 38 minutes North, 73 degrees 47 minutes West)?

    >>> lax = (radians(33 + 57 / 60.0), -radians(118 + 24 / 60.0))
    >>> jfk = (radians(40 + 38 / 60.0), -radians(73 + 47 / 60.0))

    The distance is 2143.7 nautical miles:
    >>> round(great_circle(lax, jfk)[0], 1)
    2143.7

    The initial true course out of LAX is 66 degrees:
    >>> round(degrees(great_circle(lax, jfk)[1]), 1)
    65.9

    pre:
        -pi <= lat1 <= pi
        -pi <= lat2 <= pi
        -pi <= long1 <= 2 * pi
        -pi <= long2 <= 2 * pi
    """
    # of two formula, this one is less susceptible to rounding errors
    # for short distances
    #
    lat1, long1 = pos1
    lat2, long2 = pos2
    cos_lat1 = cos(lat1)

    d = 2 * asin(sqrt(sin((lat1 - lat2) / 2.0) ** 2 + \
                      cos_lat1 * cos(lat2) * sin((long2 - long1) / 2.0) ** 2))

    # Short circuit if the same point
    if d == 0.0: return (0.0, 0.0)

    tc1 = (sin(lat2) - sin(lat1) * cos(d)) / (sin(d) * cos_lat1)
    tc1 = acos(max(min(tc1, 1.0), -1.0))

    if sin(long1 - long2) >= 0:
        tc1 = 2 * pi - tc1

    return 10800.0 * d / pi, tc1

def intersect(v1, v2):
    lat1, lon1, r1 = v1
    lat2, lon2, r2 = v2

    dst12 = 2 * asin(sqrt((sin((lat1-lat2) / 2)) ** 2 + \
            cos(lat1) * cos(lat2) * sin( (lon1-lon2) / 2) ** 2))

    if sin(lon2-lon1) < 0.0:
        crs12 = acos((sin(lat2) - sin(lat1) * cos(dst12)) / \
                (sin(dst12) * cos(lat1)))
        crs21 = 2.0 * pi - acos((sin(lat1) - sin(lat2) * cos(dst12)) / \
                (sin(dst12) * cos(lat2)))
    else:
        crs12 = 2.0 * pi - acos((sin(lat2) - sin(lat1) * cos(dst12)) / \
                (sin(dst12) * cos(lat1)))
        crs21 = acos((sin(lat1) - sin(lat2) * cos(dst12)) / \
                (sin(dst12) * cos(lat2)))

    ang1 = fmod(r1 - crs12 + pi, 2.0*pi) - pi
    ang2 = fmod(crs21 - r2 + pi, 2.0*pi) - pi

    if sin(ang1) == 0 and sin(ang2) == 0:
        raise Exception("infinity of intersections")

    if sin(ang1) * sin(ang2) < 0:
        raise Exception("intersection ambiguous")

    ang1 = abs(ang1)
    ang2 = abs(ang2)
    ang3 = acos(-cos(ang1) * cos(ang2) + sin(ang1) * sin(ang2) * cos(dst12)) 
    dst13 = atan2(sin(dst12) * sin(ang1) * sin(ang2),
                  cos(ang2) + cos(ang1) * cos(ang3))
    lat3 = asin(sin(lat1) * cos(dst13) + cos(lat1) * sin(dst13) * cos(r1))
    dlon = atan2(sin(r1) * sin(dst13) * cos(lat1),
                 cos(dst13) - sin(lat1) * sin(lat3))
    lon3 = fmod(lon1 - dlon + pi, 2*pi) - pi

    return lat3, lon3

def _test():
    import doctest, nav
    return doctest.testmod(nav)

if __name__ == '__main__':
    _test()
